package eu.gronos.kostenrechner.controller.system;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;

import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JMenuItem;
import javax.swing.JToggleButton;

import eu.gronos.kostenrechner.controller.BeschriebeneAktion;
import eu.gronos.kostenrechner.controller.ComponentBeschrifter;
import eu.gronos.kostenrechner.interfaces.ConfigOption;
import eu.gronos.kostenrechner.model.beschriftungen.BooleanConfigSchluessel;
import eu.gronos.kostenrechner.model.beschriftungen.LangBeschriftung;
import eu.gronos.kostenrechner.model.beschriftungen.LangVorsorgeBeschriftung;
import eu.gronos.kostenrechner.model.beschriftungen.NameContainer;
import eu.gronos.kostenrechner.model.beschriftungen.SchluesselBeschriftung;

/**
 * Eine {@link BeschriebeneAktion} (und damit auch {@link ActionListener} für
 * {@link AbstractButton}-Elemente, vor {@link JToggleButton} und
 * {@link JCheckBoxMenuItem}.
 * 
 * Jede Instanz korrespondiert mit einem Boolean-Wert <code>schluessel</code> in
 * der Registry im Unterast <code>"/eu/gronos/kostentenor"</code> zum Java-Ast.
 * Diese wird zunächst eingelesen. Wenn der Schlüssel <code>schluessel</code>
 * noch nicht gesetzt ist, wird der als <code>standardWert</code> übergebene
 * Wahrheitswert genommen.
 *
 * @author Peter Schuster (setrok)
 * @date 18 Feb 2019
 *
 */
public class BooleanConfigAction extends BeschriebeneAktion implements ConfigOption<Boolean> {
	/* BooleanConfigOption */
	/* ItemListener ActionListener, HorcherWandelbar */

	private static final long serialVersionUID = 6129451338244952803L;
	// private AbstractButton button;
	// private SchluesselBeschriftung beschriftung;
	private final BooleanConfigOption option;

	/**
	 * @param beschriftung eine {@link SchluesselBeschriftung}, der auch ein
	 *                     {@link BooleanConfigSchluessel},
	 *                     ({@link javax.swing.Action#NAME}),
	 *                     ({@link javax.swing.Action#SHORT_DESCRIPTION}),
	 *                     {@link javax.swing.Action#MNEMONIC_KEY},
	 *                     {@link javax.swing.Action#ACCELERATOR_KEY} und
	 *                     ({@link javax.swing.Action#ACTION_COMMAND_KEY}) entnommen
	 *                     werden.
	 */
	public BooleanConfigAction(SchluesselBeschriftung beschriftung) {
		super(beschriftung);
		option = new BooleanConfigOption((BooleanConfigSchluessel) beschriftung.getConfigSchluessel());
	}
	// super((BooleanConfigSchluessel) beschriftung.getConfigSchluessel());
	// setBeschriftung(beschriftung);
	// setButton(null);

	/**
	 * Die Methode wird aufgerufen, wenn sich der Status
	 * {@link AbstractButton#isSelected()} ändert.
	 * 
	 * Beim Klicken wird der Text des {@link AbstractButton} der Auswahl angepasst
	 * und mit {@link #inRegistrySchreiben(boolean)} der Wert auch gespeichert.
	 * 
	 * @param e ein {@link ActionEvent}
	 * 
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	@Override
	public void actionPerformed(ActionEvent e) {
		Object source = e.getSource();
		if (source instanceof AbstractButton) {
			setValueFromButton((AbstractButton) source);
		}
	}
	// AbstractButton button = (AbstractButton) source;
	// ((AbstractButton) source)

	/**
	 * Die Methode leitet die Anfrage an die {@link #option} weiter
	 * 
	 * @return den {@link BooleanConfigOption#getStandardWert()} der {@link #option}
	 *         als {@link Boolean}
	 * 
	 * @see eu.gronos.kostenrechner.interfaces.ConfigOption#getStandardWert()
	 */
	@Override
	public Boolean getStandardWert() {
		return option.getStandardWert();
	}

	/**
	 * Die Methode leitet die Anfrage an die {@link #option} weiter
	 * 
	 * @return den {@link BooleanConfigOption#getSchluessel()} der {@link #option}
	 *         als {@link Boolean}
	 * 
	 * @see eu.gronos.kostenrechner.interfaces.ConfigOption#getSchluessel()
	 */
	@Override
	public String getSchluessel() {
		return option.getSchluessel();
	}

	/**
	 * Die Methode schreibt den Wert in die Registry/Preferences und setzt das
	 * Oberflächenelement entsprechend
	 * 
	 * @param value
	 */
	public void setValue(Boolean value) {
		option.setValue(value);
		setValueToButton(value);//
	}
	// super

	/**
	 * 
	 * Die Methode liest den Wert aus der Registry (den Preferences) und setzt das
	 * Oberflächenelement entsprechend
	 * 
	 * @return value als {@link Boolean}
	 */
	public Boolean getValue() {// boolean
		Boolean value = option.getValue();
		setValueToButton(value);//
		return value;
	}
	// super

	/**
	 * Die Methode leitet den Aufruf an {@link #option} weiter
	 * 
	 * @see eu.gronos.kostenrechner.interfaces.ConfigOption#reset()
	 */
	@Override
	public void reset() {
		option.reset();
	}

	/**
	 * Die Methode baut einen {@link JToggleButton} mit dem {@link ItemListener} und
	 * setzt zusätzlich {@link JToggleButton#getName()} auf die
	 * {@link NameContainer#getId()}.
	 * 
	 * @return den erstellten {@link JToggleButton} als {@link AbstractButton}
	 * 
	 */
	public JToggleButton toJToggleButton() {
		return (JToggleButton) toButton(new JToggleButton());
	}
	// @Override
	// AbstractButton button = new JToggleButton();
	// * @see eu.gronos.kostenrechner.interfaces.HorcherWandelbar#toButton()

	/**
	 * Die Methode versieht einen {@link AbstractButton} oder ein abgeleitetes
	 * Objekt mit dem {@link ItemListener} und setzt zusätzlich
	 * {@link AbstractButton#getName()} auf die {@link NameContainer#getId()}.
	 * 
	 * @param button einen {@link AbstractButton}
	 * @return den präparierten {@link AbstractButton}
	 * 
	 * @see eu.gronos.kostenrechner.interfaces.HorcherWandelbar#toButton(javax.swing.AbstractButton)
	 */
	@Override
	public AbstractButton toButton(AbstractButton button) {
		button.setAction(this);
		getValue();
		setButtonText();
		return button;
	}
	// this.setButton(button);
	// button
	// button.addActionListener(this);
	// button.addItemListener(this);
	// button.setAction(this);
	// actionPerformed(new ActionEvent(button, ActionEvent.ACTION_FIRST,
	// getBeschriftung().getActionCommandKey()));
	// berechnungsMethodeAction.actionPerformed(new
	// ActionEvent(btnBerechnungsMethode, ActionEvent.ACTION_FIRST,
	// btnBerechnungsMethode.getActionCommand()));

	/**
	 * Die Methode baut ein {@link JMenuItem} mit dem {@link ItemListener} und setzt
	 * zusätzlich {@link JMenuItem#getName()} auf die {@link NameContainer#getId()}.
	 * 
	 * @return das erstellte {@link JMenuItem}
	 * 
	 * @see eu.gronos.kostenrechner.interfaces.HorcherWandelbar#toMenuItem()
	 */
	@Override
	public JMenuItem toMenuItem() {
		JCheckBoxMenuItem item = new JCheckBoxMenuItem();
		return toMenuItem(item);
	}

	/**
	 * Die Methode versieht ein {@link JMenuItem} oder ein abgeleitetes Objekt mit
	 * dem {@link ItemListener} und setzt zusätzlich {@link JMenuItem#getName()} auf
	 * die {@link NameContainer#getId()}.
	 * 
	 * @param item ein {@link JMenuItem}
	 * @return das präparierte {@link JMenuItem}
	 * 
	 * @see eu.gronos.kostenrechner.interfaces.HorcherWandelbar#toMenuItem(javax.swing.JMenuItem)
	 */
	@Override
	public JMenuItem toMenuItem(JMenuItem item) {
		ComponentBeschrifter beschrifter = new ComponentBeschrifter();
		beschrifter.beschrifte(item, getBeschriftung());
		item.setAction(this);
		getValue();
		setButtonText();
		return item;
	}
	// setButton(item);
	// item.addActionListener(this);
	// .applyTo
	// item.addItemListener(this);
	// item.setAction(this);
	// button

	/**
	 * @return gibt {@link #beschriftung} als {@link LangBeschriftung} zurück, der
	 *         auch ({@link javax.swing.Action#NAME}),
	 *         ({@link javax.swing.Action#SHORT_DESCRIPTION}),
	 *         {@link javax.swing.Action#MNEMONIC_KEY},
	 *         {@link javax.swing.Action#ACCELERATOR_KEY} und
	 *         ({@link javax.swing.Action#ACTION_COMMAND_KEY}) entnommen werden.
	 */
	@Override
	public SchluesselBeschriftung getBeschriftung() {
		return (SchluesselBeschriftung) super.getBeschriftung();
		// return beschriftung;
	}

	/*
	 * @param beschriftung d. {@link #beschriftung}, d. gesetzt werden soll als
	 * {@link LangBeschriftung}.
	 */
	// public void setBeschriftung(SchluesselBeschriftung beschriftung) {
	// this.beschriftung = beschriftung;
	// }

	/**
	 * Beim Klicken wird der Text des {@link AbstractButton} aus den
	 * {@link LangVorsorgeBeschriftung#getAlternativeTitles()} der Auswahl
	 * angepasst, bei {@link Boolean#TRUE} der index 1, bei {@link Boolean#FALSE}
	 * der index 0.
	 * 
	 * Bei formattierten Beschriftungen muss das die abgeleitete Klasse übernehmen.
	 * 
	 */
	protected void setButtonText() {
		int index = valueToIndex();
		ComponentBeschrifter beschrifter = new ComponentBeschrifter();
		beschrifter.changeTo(getBeschriftung(), index);
		beschrifter.beschrifte(this, getBeschriftung());
	}
	// if (getBeschriftung().isFormatted()) {
	// Formattierung muss jemand anderes übernehmen
	// return;
	// }
	// AbstractButton button @param button der {@link #toButton()}
	// button//button
	// getBeschriftung() instanceof SchluesselBeschriftung &&
	// 0 : 1;

	/**
	 * Die Methode gibt statt des {@link Boolean} aus {@link #getValue()} einen
	 * <code>int</code> zurück, um einen index für die
	 * {@link LangVorsorgeBeschriftung} hat.
	 * 
	 * @return 1 bei <code>true</code>, 0 bei <code>false</code>
	 */
	private int valueToIndex() {
		return getValue().booleanValue() ? 1 : 0;
	}
	// AbstractButton button
	// button.isSelected()

	/*
	 * @param button der {@link AbstractButton}, an dem die Option an der Oberfläche
	 * geändert werden kann und für den die Instanz als {@link ItemListener}
	 * eingerichtet werden soll.
	 */
	// private void setButton(AbstractButton button) {
	// this.button = button;
	// }
	// AbstractButton button = (AbstractButton) source;
	// ((AbstractButton) source)

	/**
	 * Die Methode liest den Wert vom {@link AbstractButton#isSelected()} und
	 * schreibt ihn in die {@link #option}
	 * 
	 * @param button den {@link AbstractButton#isSelected()}
	 */
	private void setValueFromButton(AbstractButton button) {
		boolean value = button.isSelected();
		option.inRegistrySchreiben(value);
		setButtonText();// button
	}

	/**
	 * Die Methode setzt das Oberflächenelement über
	 * {@link Action#SELECTED_KEY }entsprechend dem {@link #getValue()}.
	 * 
	 * @param value <code>true</code> if the button is selected, otherwise false
	 */
	private void setValueToButton(boolean value) {
		// Der value darf hier nicht aus #getValue gezogen werden, sonst Endlosschleife
		putValue(Action.SELECTED_KEY, value);
	}
	// if (button != null) {
	// button.setSelected(value);
	// }
	// getValue()

}

/*
 * @param schluessel der Schlüssel im Unterast
 * <code>"/eu/gronos/kostentenor"</code> (String)
 * 
 * @param standardWert der Standard-Wert, falls der Schlüssel noch nicht
 * angelegt ist. (boolean)
 * 
 * @param button der {@link AbstractButton}, an dem die Option an der Oberfläche
 * geändert werden kann und für den die Instanz als {@link ItemListener}
 * eingerichtet werden soll.
 */
//public BooleanConfigOptionHorcher(BooleanConfigSchluessel parameter, AbstractButton button) {
//	super(parameter);
//	setButton(button);
//	if (button.getAction() != null && button.getAction() instanceof BeschriebeneAktion) {
//		setBeschriftung(((BeschriebeneAktion) button.getAction()).getBeschriftung());
//	}
//	getValue();
//}
/*
 * String schluessel, boolean standardWert new
 * BooleanConfigSchluessel(schluessel, standardWert)
 */

/*
 * 
 * @return gibt den {@link AbstractButton}, an dem die Option an der Oberfläche
 * geändert werden kann und für den die Instanz als {@link ItemListener}
 * eingerichtet werden soll.
 */
//private AbstractButton getButton() {
//	return button;
//}

/*
 * @param e ein {@link ItemEvent}
 * 
 * @see java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
 */
//@Override
//public void itemStateChanged(ItemEvent e) {
//	Object source = e.getSource();
//	if (source instanceof AbstractButton) {
//		boolean value = ((AbstractButton) source).isSelected();
//		inRegistrySchreiben(value);
//	}
//}
// public BooleanConfigOptionHorcher(BooleanConfigSchluessel parameter,
// LangBeschriftung beschriftung) {
// super(parameter);
// setBeschriftung(beschriftung);
// setButton(null);
// }
// String schluessel, boolean standardWert
// new BooleanConfigSchluessel(schluessel, standardWert)

/*
 * Die Methode setzt wieder den Standardwert in die Registry/Preferences.
 * 
 * @see eu.gronos.kostenrechner.controller.system.BooleanConfigOption#reset()
 */
// public void reset() {
// setValue(getStandardWert());
// }