/*
 * GebuehrenVerzeichnis.java
 * eu.gronos.kostenrechner.model.gebuehren (Kostenrechner)
 */
package eu.gronos.kostenrechner.model.gebuehren;

import java.util.ArrayList;

import eu.gronos.kostenrechner.model.tenordaten.TenorDatenContainer;

/**
 * Von {@link ArrayList} abgeleitete und aus {@link GebuehrenVerzeichnis}
 * hervorgegangene Klasse zur Übergabe von {@link GebuehrenTatbestand}en im
 * {@link TenorDatenContainer}, die auch Methoden zur Berechnung enthält,
 * nämlich die statische Methode
 * {@link #errechneGebuehr(GebuehrenTatbestand, long)} und für die Summe
 * {@link #errechneGebuehrenSumme(long)} sowie
 * {@link #errechneGebuehrenSummeMitSteuer(long)}.
 *
 * @author Peter Schuster (setrok)
 * @date 23.06.2019
 *
 */
public class GebuehrenAuflistung extends ArrayList<GebuehrenTatbestand> {

	private static final long serialVersionUID = 6564487211842926610L;

	public GebuehrenAuflistung() {
		super();
	}
	

	/**
	 * Die Methode dient dazu, die passende Gebühr oder Auslage zu berechnen.
	 * 
	 * @param gt         einen GebuehrenSatzTatbestand oder AuslagenTatbestand,
	 *                   andere Ableger von GebuehrenTatbestand werden ignoriert.
	 * @param streitwert Streitwert (wird bei Auslagen ignoriert)
	 * @return die Höhe der Gebühr (dann errechnet nach 1,0 Gebühr für den
	 *         streitwert, multipliziert mit dem Gebührensatz nach gt) bzw. Auslage
	 */
	public static double errechneGebuehr(GebuehrenTatbestand gt, long streitwert) {
		if (gt != null && gt instanceof AuslagenTatbestand)
			return ((AuslagenTatbestand) gt).getBetrag();
		else if (gt != null && gt instanceof PauschalTatbestand)
			return ((PauschalTatbestand) gt).getBetrag();
		else if (gt != null && gt instanceof GebuehrenSatzTatbestand)
			return GebuehrenAuflistung.findeGebuehrenTabelle(gt).errechneGebuehr(streitwert,
					((GebuehrenSatzTatbestand) gt).getSatz());
		else
			return 0.0;
	}

	/**
	 * Die Methode soll die Summe aller Gebühren liefern. Mehrwertsteuer wird
	 * ignoriert.
	 * 
	 * @param streitwert ein long mit dem Streitwert
	 * @return ein double mit der Summe der Gebühren für diesen Streitwert
	 */
	public double errechneGebuehrenSumme(long streitwert) {
		double summe = 0;
		for (GebuehrenTatbestand gt : this) {
			/* summe um die Gebühr erhöhen */
			summe += GebuehrenAuflistung.errechneGebuehr(gt, streitwert);
		}
		return summe;
	}

	/**
	 * Die Methode soll die Summe aller Gebühren der Tabelle liefern und auch die
	 * Umsatzsteuer ausrechnen, wenn sie im {@link GebuehrenAuflistung} vorkommen.
	 * 
	 * @param streitwert ein long mit dem Streitwert
	 * @return ein double mit der Summe der Gebühren für diesen Streitwert
	 */
	public double errechneGebuehrenSummeMitSteuer(long streitwert) {
		double summe = 0;
		double rvgsumme = 0;
		for (GebuehrenTatbestand gt : this) {
			/*
			 * errechneGebuehr ignoriert MehrwertsteuerTatbestände; diese muss deshalb extra
			 * gerechnet werden.
			 */
			if (gt instanceof MehrwertsteuerTatbestand) {
				MehrwertsteuerTatbestand mwst = (MehrwertsteuerTatbestand) gt;
				summe += mwst.errechneSteuer(rvgsumme);
				/*
				 * Wenn die Gebühren isoliert werden, die RVG-Summe wieder auf 0 setzen.
				 */
				if (mwst.isIsoliert())
					rvgsumme = 0.0;
			} else {
				/* summe um die Gebühr erhöhen */
				summe += GebuehrenAuflistung.errechneGebuehr(gt, streitwert);
				/*
				 * rvgsumme nur erhöhen, wenn RVG-Gebühr (wird für MwSt gebraucht).
				 */
				if (gt.getGebuehrenKlasse() == AnwaltsGebuehrenTabelle.class)
					rvgsumme += GebuehrenAuflistung.errechneGebuehr(gt, streitwert);
			}
		}
		return summe;
	}

	/**
	 * Die Methode findeGebuehrenTabelle findet heraus, ob ein Gebührentatbestand
	 * zum RVG oder GKG gehört und gibt eine Instanz dieser Tabelle zurück.
	 * 
	 * @param gt der GebuehrenTatbestand
	 * @return die zum GebuehrenTatbestand gehörende GebuehrenTabelle
	 */
	private static GebuehrenTabelle findeGebuehrenTabelle(GebuehrenTatbestand gt) {
		if (AnwaltsGebuehrenTabelle.class.equals(gt.getGebuehrenKlasse()))
			return GebuehrenGesetzesSammlung.getRvg();
		else if (GerichtsGebuehrenTabelle.class.equals(gt.getGebuehrenKlasse()))
			return GebuehrenGesetzesSammlung.getGkg();
		return null;
	}

}
//	private String title = new String();
//
//	/**
//	 * Die Methode ist ein modifizierter {@link #get(int)}-Getter
//	 * 
//	 * @param vorschriften Auflistung der
//	 * @return alle Elemente, die {@link GebuehrenTatbestand#getVorschrift()}
//	 *         enthalten
//	 * @throws IllegalArgumentException, wenn eine der vorschriften nicht gefunden
//	 *                                   wird
//	 */
//	public GebuehrenAuflistung getAll(String... vorschriften) throws IllegalArgumentException {
//		GebuehrenAuflistung verzeichnis = new GebuehrenAuflistung();
//		boolean gefunden = false;
//		for (String vorschrift : vorschriften) {
//			gefunden = false;
//			for (GebuehrenTatbestand gt : this) {
//				if (vorschrift != null && gt != null && vorschrift.equals(gt.getVorschrift())) {
//					gefunden = true;
//					verzeichnis.add(gt);
//				}
//			}
//			if (!gefunden) {
//				throw new IllegalArgumentException(
//						String.format("Gebührentatbestand nicht im Verzeichnis: %s!", vorschrift));
//			}
//		}
//
//		return verzeichnis;
//	}
//
//	/**
//	 * Die Methode fügt nur die Elemente aus dem Verzeichnis hinzu, die den
//	 * Suchbegriff {@link String} enthalten. Groß-/Kleinschreibung wird ignoriert.
//	 * 
//	 * @param term ein {@link String} mit dem Suchbegriff
//	 * @return {@link GebuehrenAuflistung}
//	 */
//	public GebuehrenAuflistung filter(String term) {
//		GebuehrenAuflistung verzeichnis = new GebuehrenAuflistung();
//		for (GebuehrenTatbestand gt : this) {
//			if (gt != null && gt.toString().toLowerCase().contains(term.toLowerCase())) {
//				verzeichnis.add(gt);
//			}
//		}
//		return verzeichnis;
//	}
//
//	/**
//	 * @return gibt {@link #title} als {@link String} zurück.
//	 */
//	@XmlAttribute(name = "title")
//	public String getTitle() {
//		return title;
//	}
//
//	/**
//	 * @param title d. {@link #title}, d. gesetzt werden soll als {@link String}.
//	 */
//	public void setTitle(String title) {
//		this.title = title;
//	}
	/*
	 * @param tatbestaende ein {@link GebuehrenVerzeichnis} mit den angefallenen
	 * Gebührentatbeständen. public static double
	 * errechneGebuehrenSumme(GebuehrenVerzeichnis tatbestaende, long streitwert) {
	 * for (GebuehrenTatbestand gt : tatbestaende) {
	 * 
	 * }
	 */